@@ -1,5 +1,7 @@
 # Make the more user-facing classes and objects accessible at the top-level (e.g. schem.levels.levels => schem.levels)
+from ._version import __version__
 from .game import run, validate
+from .precognition import is_precognitive
 from .level import Level
 from .levels import levels  # This name overlap mostly hides the levels module which I prefer anyway
 from .terrains import terrains  # Ditto
