@@ -6,10 +6,12 @@ from typing import Optional
 from .level import Level
 from .levels import levels as built_in_levels, defense_names, resnet_ids
 from .exceptions import ScoreError
-from .solution import Score, Solution, DebugOptions
+from .precognition import is_precognitive
+from .solution import Solution, DebugOptions
 
 
-def run(soln_str: str, level_code=None, level_codes=None, max_cycles=None, return_json=False, verbose=False,
+def run(soln_str: str, level_code=None, level_codes=None, max_cycles=None, validate_expected_score=False,
+        return_json=False, check_precog=False, max_precog_check_cycles=None, verbose=False, stderr_on_precog=False,
         debug: Optional[DebugOptions] = False):
     """Wrapper for Solution.run which identifies the level to run in as needed, based on the solution metadata.
 
@@ -33,8 +35,10 @@ def run(soln_str: str, level_code=None, level_codes=None, max_cycles=None, retur
             Note that while some built-in game levels have duplicate titles, currently all of these have conflicting
             level features (bonders etc.), and therefore a solution can't successfully load into / solve multiple
             built-in levels.
-        max_cycles: Maximum cycle count to run to. Default double the expected cycle count in the solution metadata,
+        max_cycles: Maximum cycle count to run to. Default 1.1x the expected cycle count in the solution metadata,
             or 1,000,000 cycles if no expected score (use math.inf if you don't fear infinite loop solutions).
+        validate_expected_score: If True, raise an exception if the solution metadata's expected score is missing or
+                                 doesn't match the run result. Default False.
         return_json: If True, instead of a Score return a dict including the usual score fields, but also the level
             title, ResearchNet volume-issue-puzzle tuple (None if not a ResearchNet level), and solution author/title.
 
@@ -43,12 +47,34 @@ def run(soln_str: str, level_code=None, level_codes=None, max_cycles=None, retur
             return the metadata of the first level that the solution could be successfully imported into.
 
             Default False.
+        check_precog: If True, do additional runs on the solution to check if it fits the current community definition
+            of a precognitive solution. Currently only useful if return_json is present, in which case a 'precog' field
+            will be included. See `schem.precognition.is_precognitive` for more info and a direct API.
+        max_precog_check_cycles: The maximum total cycle count that may be used by all precognition-check runs; if this
+            value is exceeded before sufficient confidence in an answer is obtained, a TimeoutError is raised, or in the
+            case of return_json, the 'precog' field is set to None.
+            Default 2,000,000 cycles (this is sufficient for basically any sub-10k solution).
         verbose: If True, print warnings if there is not exactly one level with title matching the solution metadata.
                  Default False.
+        stderr_on_precog: If True, when a solution is precognitive, print an explanation of why to STDERR.
+                          Can be enabled independently of `verbose`. Default False.
         debug: Print an updating view of the solution while running; see DebugOptions. Default False.
     """
     assert level_code is None or level_codes is None, "Only one of level_code or level_codes may be specified"
-    level_name, _, expected_score, _ = Solution.parse_metadata(soln_str)
+    level_name, author, expected_score, soln_name = Solution.parse_metadata(soln_str)
+    soln_descr = Solution.describe(level_name, author, expected_score, soln_name)
+
+    if validate_expected_score:
+        if expected_score is None:
+            raise ValueError("Validation requires a valid expected score in the first solution line (currently 0-0-0);"
+                             + " please update it or use run() without validation instead.")
+
+        if max_cycles is not None:
+            if expected_score.cycles > max_cycles:
+                raise ValueError(f"{soln_descr}: Cannot validate; expected cycles > max cycles ({max_cycles})")
+
+            # If validating, limit the max cycles based on the expected score, to save time
+            max_cycles = min(max_cycles, int(expected_score.cycles * 1.1))
 
     # Convert level_code convenience arg to same format as level_codes
     if level_code is not None:
@@ -97,7 +123,13 @@ def run(soln_str: str, level_code=None, level_codes=None, max_cycles=None, retur
     for level, resnet_id in zip(matching_levels, matching_resnet_ids):
         try:
             solution = Solution(level=level, soln_export_str=soln_str)
-            cur_score = cycles, reactors, symbols = solution.run(max_cycles=max_cycles, debug=debug)
+            score = cycles, reactors, symbols = solution.run(max_cycles=max_cycles, debug=debug)
+
+            # Avoid doing e.g. precog checks if expected score validation failed
+            # Note that the ScoreError will be caught and stored by the except clause so we can remember this score
+            # while still continuing to check any other matching levels
+            if validate_expected_score and score != expected_score:
+                raise ScoreError(f"{soln_descr}: Expected score {expected_score} but got {score}")
 
             run_data = {'level_name': level.name,
                         'resnet_id': resnet_id,
@@ -107,24 +139,44 @@ def run(soln_str: str, level_code=None, level_codes=None, max_cycles=None, retur
                         'author': solution.author,
                         'solution_name': solution.name}
 
-            # Return the successful run if there was no expected score or it matched the expected cycle count
-            if expected_score is None or cur_score.cycles == expected_score.cycles:
-                return run_data if return_json else cur_score
+            if check_precog:
+                try:
+                    run_data['precog'] = is_precognitive(solution, max_cycles=max_cycles,
+                                                         max_total_cycles=max_precog_check_cycles,
+                                                         just_run_cycle_count=score.cycles,
+                                                         verbose=verbose,
+                                                         stderr_on_precog=stderr_on_precog)
+                except TimeoutError:
+                    # If using --json mode, store None for the field instead of propagating any timeout error
+                    if return_json:
+                        run_data['precog'] = None
+                    else:
+                        raise
+
+            # Return the successful run immediately if there was no expected score or it matched
+            if expected_score is None or score.cycles == expected_score.cycles:
+                ret_val = run_data if return_json else score
+                break
 
             # Preserve the first successful score (in case the expected score is never found)
             if ret_val is None:
-                ret_val = run_data if return_json else cur_score
+                ret_val = run_data if return_json else score
         except Exception as e:
             exceptions.append(e)
 
-    # Raise the first error if no successful run was found
+    # Return the first successful run or raise the first error
     if ret_val is not None:
+        if verbose:
+            if validate_expected_score:
+                print(f"Validated {soln_descr}")
+
         return ret_val
     else:
         raise exceptions[0]
 
 
-def validate(soln_str: str, level_code=None, level_codes=None, max_cycles=None, return_json=False, verbose=False,
+def validate(soln_str: str, level_code=None, level_codes=None, max_cycles=None, return_json=False, check_precog=False,
+             max_precog_check_cycles=None, verbose=False, stderr_on_precog=False,
              debug: Optional[DebugOptions] = False):
     """Sibling of Solution.validate which identifies the level to run in as needed, based on the solution metadata.
 
@@ -145,7 +197,7 @@ def validate(soln_str: str, level_code=None, level_codes=None, max_cycles=None,
             Note that while some built-in game levels have duplicate titles, currently all of these have conflicting
             level features (bonders etc.), and therefore a solution can't successfully load into / solve multiple
             built-in levels.
-        max_cycles: Maximum cycle count to run to. Default double the expected cycle count in the solution metadata.
+        max_cycles: Maximum cycle count to run to. Default 1.1x the expected cycle count in the solution metadata.
         return_json: If True, instead of a Score return a dict including the usual score fields, but also the level
             title, ResearchNet volume-issue-puzzle tuple (None if not a ResearchNet level), and solution author/title.
 
@@ -154,39 +206,23 @@ def validate(soln_str: str, level_code=None, level_codes=None, max_cycles=None,
             return the metadata of the first level that the solution could be successfully imported into.
 
             Default False.
+        check_precog: If True, do additional runs on the solution to check if it fits the current community definition
+            of a precognitive solution. Currently only useful if return_json is present, in which case a 'precog' field
+            will be included. See `schem.precognition.is_precognitive` for more info and a direct API.
+        max_precog_check_cycles: The maximum total cycle count that may be used by all precognition-check runs; if this
+            value is exceeded before sufficient confidence in an answer is obtained, a TimeoutError is raised, or in the
+            case of return_json, the 'precog' field is set to None.
+            Default 2,000,000 cycles (this is sufficient for basically any sub-10k solution).
         verbose: If True, print warnings if there is not exactly one level with title matching the solution metadata.
                  Default False.
+        stderr_on_precog: If True, when a solution is precognitive, print an explanation of why to STDERR.
+                          Can be enabled independently of `verbose`. Default False.
         debug: Print an updating view of the solution while running; see DebugOptions. Default False.
-
     """
-    level_name, author, expected_score, soln_name = Solution.parse_metadata(soln_str)
-    if expected_score is None:
-        raise ValueError("validate() requires a valid expected score in the first solution line (currently 0-0-0);"
-                         + " please update it or use run() instead.")
-
-    # TODO: Should use level_code's name if conflicting
-    soln_descr = Solution.describe(level_name, author, expected_score, soln_name)
-
-    if max_cycles is not None and expected_score.cycles > max_cycles:
-        raise ValueError(f"{soln_descr}: Cannot validate; expected cycles > max cycles ({max_cycles})")
-
     ret_val = run(soln_str, level_code=level_code, level_codes=level_codes, max_cycles=max_cycles,
-                  return_json=return_json, verbose=verbose, debug=debug)
-
-    if not return_json:
-        score = ret_val
-    # run(return_json=True) only raises an error on solution load errors; runtime errors just set cycles to None
-    elif ret_val['cycles'] is None:
-        raise Exception(f"{soln_descr}: Expected score {expected_score} but solution crashed or timed out.")
-    else:
-        score = Score(ret_val['cycles'], ret_val['reactors'], ret_val['symbols'])
-
-    # Validate the score
-    if score != expected_score:
-        raise ScoreError(f"{soln_descr}: Expected score {expected_score} but got {score}")
-
-    if verbose:
-        print(f"Validated {soln_descr}")
+                  validate_expected_score=True, return_json=return_json, check_precog=check_precog,
+                  max_precog_check_cycles=max_precog_check_cycles,
+                  verbose=verbose, stderr_on_precog=stderr_on_precog, debug=debug)
 
     if return_json:
         return ret_val
